@@ -18,6 +18,8 @@ from linearmodels import PanelOLS
 from linearmodels.panel.model import PanelFormulaParser
 from linearmodels.panel.utility import check_absorbed, not_absorbed, AbsorbingEffectError
 
+from functools import reduce
+
 class UnequalRowException(Exception):
     pass
 
@@ -64,9 +66,13 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
                  drop_absorbed=False,
                  check_rank=True,
                  singletons = True,
+                 check_absorbed=True,
                  **kwargs):
         
         print("Initializing...")    
+        
+        
+        self.check_absorbed= check_absorbed
 
         if classifier_ind is None:
             self._group = endog.index.get_level_values(2).name
@@ -156,6 +162,10 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
             warnings.warn("Warning: observations across regimes is not the same, switching to slower method...")
             self._equal_regime=False
             
+        self._get_vars_index = []
+        self._get_vars = [] 
+        self.entity, self.time = self.endog_cat.index.names
+            
 
         
     @classmethod
@@ -169,6 +179,7 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
         drop_absorbed=False,
         check_rank=True,
         singletons = True,
+        check_absorbed=True,
         **kwargs
     ):
         
@@ -195,6 +206,7 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
             drop_absorbed=drop_absorbed,
             check_rank=check_rank,
             singletons = singletons,
+            check_absorbed=check_absorbed,
             **kwargs
         )
                 
@@ -202,45 +214,61 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
         
         return mod
     
-    @functools.lru_cache(maxsize=128)
-    def _regime_list(self, lm):            
-        
-        if self.demean is not None:
-            return np.array([PanelData(group.drop(columns=self._group), 
-                                       convert_dummies=False).demean(group=self.demean, low_memory=False).values2d \
-                for _, group in getattr(self, lm).groupby(self._group)])
+    def demean_data(self, data):
+        if self.demean=='entity':
+            return data.groupby(self.entity).transform(lambda x: x - x.mean())
+        elif self.demean == 'time':
+            return data.groupby(self.time).transform(lambda x: x - x.mean())
+        elif self.demean == 'both':
+            return data.groupby(self.entity).transform(lambda x: x - x.mean()).groupby(self.time).transform(lambda x: x - x.mean())
         else:
-            return np.array([PanelData(group.drop(columns=self._group), convert_dummies=False).values2d \
-                for _, group in getattr(self, lm).groupby(self._group)])
+            return data
+        
+    @functools.lru_cache(maxsize=128)
+    def _regime_list(self, lm):       
+        return np.array([group.drop(columns=self._group).pipe(self.demean_data).values \
+            for _, group in getattr(self, lm).groupby(self._group)])
               
     @functools.cached_property
     def X(self):
         X = self._regime_list('exog_full')
-        
-        # if not self._equal_regime:
-        #     print("Regimes not equal... Will merge by shortest dataset")
-        #     # Get smallest dataset
-        #     minidx = np.argmin([i.dataframe.shape[0] for i in X])
-        #     index_to_match = X[minidx].index
             
-        #     X = np.array([i.dataframe.query("index in @index_to_match") for i in X])
+        if self.check_absorbed:
+            # (regimes, rows, columns)
             
-        # (regimes, rows, columns)
-        for i, regime in enumerate(X):
-            # try:
-            check_absorbed(regime, self.exog_full.columns.tolist())
-            # TODO: figure out a way to drop absorbed variables
-            # except AbsorbingEffectError as e:
-            #     if self._drop_absorbed:
-            #         X[i, :, :] = not_absorbed(regime, self.p_model._constant, 
-            #                                   self.p_model._constant_index)
+            exception_flag = False
+            for i, regime in enumerate(X):
+                try:
+                    check_absorbed(regime, self.exog_full.columns.tolist())
+                except AbsorbingEffectError as e:
+                    get_vars = {i for i in self.exog_full.columns.tolist() if i in str(e)}
+                    get_vars.discard("Intercept")
+                    
+                    print(f"""Multicollinearity Alert!
+                    For regime: {list(self.classifier_map.values())[i]}
+                    
+                    {get_vars}
+                            
+                    """)
+                    if self._drop_absorbed:
+                        self._get_vars_index.append(self.exog_full.columns.get_indexer(get_vars))
+                        self._get_vars.append(get_vars)
+                    else:
+                        exception_flag = True
+            if exception_flag:
+                raise AbsorbingEffectError
             
+            if self._get_vars and reduce(lambda x, y: x.intersection(y), self._get_vars):
+                raise AbsorbingEffectError(f"""
+                                            Common Absorbed Variables across regimes:
+                                            
+                                            {reduce(lambda x, y: x.intersection(y), self._get_vars)}
+                                            """)
+                
         return X
     
     @functools.cached_property
     def y(self):
-        # y = np.array([i.values2d for i in self._regime_list('endog_full')])
-
         return self._regime_list('endog_full')
     
     @functools.cached_property
@@ -277,7 +305,7 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
         
         return p_list
 
-    def _ll(self, sigma_vec, beta_vec):
+    def _ll(self, params):
                 
         
         if self.n_regimes == 4:
@@ -415,28 +443,35 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
                     ],
                 ]
             )
-            
-        # if self._equal_regime:
-        rnl = norm.pdf(np.squeeze(self.y), self.X@beta_vec, scale=sigma_vec).T
         
+        # generate X so that we can for absorbed variables
+        X = self.X
+        
+        results_df = pd.Series(params, index=self._param_order)
+        
+        beta_df = results_df.loc[results_df.index != 'sigma']
+        sigma = sigma = results_df.loc[results_df.index=='sigma'].values[0]
+            
+        if self._drop_absorbed and self._get_vars_index:
+            Xb = np.empty((self.n_regimes, self.exog.shape[0]))
+            # subset X and multiply by new beta vector
+            for i, (x, no_retain, no_retain_names) in enumerate(zip(X, self._get_vars_index, self._get_vars)):
+                beta_vec = beta_df.loc[lambda df: ~df.index.isin(no_retain_names)].values
+                Xb[i, :] = np.delete(x, no_retain, axis=1) @ beta_vec
+        else:
+            beta_vec = beta_df.values
+            Xb = X @ beta_vec
+        
+        # loc = (np.squeeze(self.y) - Xb)/sigma
+        # rnl = (norm._pdf(loc)/sigma).T
+        rnl = norm.pdf(np.squeeze(self.y), Xb, scale=np.abs(sigma)).T
+
         # get row_maxes for probabilities
         row_maxes = self.p.max(axis=1, keepdims=True)
         class_ind = np.where(self.p==row_maxes, self.p, 0)
     
         return np.log((rnl*(class_ind@cm.T)).sum(axis=1))
-        # else:
-        #     likelihood = []
-        #     class_ind = [np.where(p.values==p.values.max(axis=1, keepdims=True), p, 0) for p in self.p_slow]
-            
-        #     def compute_likelihood(z):
-        #         i, (x,y,c) = z
-        #         r = norm.pdf(np.squeeze(y), np.squeeze(x@beta_vec), scale=sigma_vec)
-        #         return r*(c@cm.T)[:, i]  #@cm.T
-            
-        #     # likelihood = np.array(Parallel(n_jobs=10, verbose=10)(delayed(compute_likelihood)(z) for z in zip(self.X, self.y, self.p_slow)))
-        #     likelihood = np.array([compute_likelihood(z) for z in enumerate(zip(self.X, self.y, class_ind))])
-            
-        #     return np.log(likelihood.sum(axis=1))
+
 
     def _params_to_ll(self, params):
 
@@ -457,10 +492,10 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
 
         """
 
-        beta_vec, sigma_vec = self._params_to_ll(params)
+        # beta_vec, sigma_vec = self._params_to_ll(params)
 
         ll = self._ll(
-            beta_vec=beta_vec, sigma_vec=sigma_vec,
+            params=params
         )
 
         return -ll
@@ -476,41 +511,42 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
 
         return bounds
     
-    def _start_params(self, show_ols = False):
-        """Runs OLS regression to get start params for MLE coefficients
-        """
+    def _start_params(self, show_ols = False, cluster_var=None):
+        """Runs OLS regression to get start params for MLE coefficients"""
                 
         print("Creating starting values...")
 
-        
-        res = self.p_model.fit()
+        if cluster_var is None:
+            res = self.p_model.fit()
+        else:
+            res = self.p_model.fit(cov_type="clustered", clusters=cluster_var)
         summary = res.summary
         
         # Now get sigma
         sigma = self.endog.std()
+        
+        results_df = pd.concat([res.params, 
+                                pd.DataFrame([sigma], index=['sigma'])])
 
         if show_ols:
             print(summary)          
         
-        return np.append(res.params.values, sigma)
+        return results_df, res
     
     def fit(
         self,
         method=None,
-        start_params=None,
         maxiter=10000,
         maxfun=5000,
         sigma_bound=None,
         cluster_var=None,
-        show_ols=True,
+        show_ols=False,
         **kwds,
     ):
 
-        # self.apply_exog_names(regime_impact_names=regime_impact_names)
-
-
-        if start_params is None:
-            start_params = self._start_params(show_ols=show_ols)
+        results_df, pols_res = self._start_params(show_ols=show_ols, cluster_var=cluster_var)
+        
+        self._param_order = results_df.index
         
         if kwds.get('cov_type') is None:
             cov_type='cluster'
@@ -524,16 +560,17 @@ class DriverSpecificProbUberMLE(GenericLikelihoodModel):
         print("Optimizing...")
         optimize = super().fit(
             method=method,
-            start_params=start_params,
+            start_params=results_df.values,
             maxiter=maxiter,
             maxfun=maxfun,
             eps=1e-08,
             ftol=1e-10,
-            bounds=self.bounds(num_exog = len(start_params), sigma_bound=sigma_bound),
+            bounds=self.bounds(num_exog = len(results_df), sigma_bound=sigma_bound),
             cov_type= cov_type,
             cov_kwds = cov_kwds,
             use_t=True,
             **kwds
         )
-
-        return optimize
\ No newline at end of file
+        
+        return optimize, pols_res
+        
